from interactions.ext import Base, Version
from .extension import Persistence

version = Version(version="0.1.0")

base = Base(
    name="Persistence",
    version=version,
    link=f"https://github.com/dworv/interactions-persistence",
    description="An extension to add simple custom_id encoding to interactions.py.",
    long_description=".",  # TODO
)

def setup(bot):
    return Persistence(bot)


from interactions import Client


def persistent_component(bot: Client, tag: str):
    def inner(coro):
        bot.event(coro, name="component_persistence_" + tag)
        return coro

    return inner


def persistent_modal(bot: Client, tag: str):
    def inner(coro):
        bot.event(coro, name="modal_persistence_" + tag)
        return coro
    
    return inner


def extension_persistent_component(tag: str):
    def inner(coro):
        coro.__persistence_type__ = "component"
        coro.__persistence_tag__ = tag
        return coro

    return inner


def extension_persistent_modal(tag: str):
    def inner(coro):
        coro.__persistence_type__ = "modal"
        coro.__persistence_tag__ = tag
        return coro

    return inner



from .base import base, version, setup
from .parse import PersistentCustomID, persistent_custom_id, ParseError
from .extension import PersistenceExtension
from .decor import extension_persistent_component, extension_persistent_modal


def persistent_custom_id(tag: str, package: "SupportsRepr"):
    return str(PersistentCustomID.new(tag, package))

class ParseError(BaseException):
    pass

class PersistentCustomID:
    def __init__(self, tag: str, payload: str):  # noqa
        self.tag: str = tag
        self.payload = payload

    @property
    def packed(self) -> str:
        return f"persistence_{self.tag}:{self.payload}"

    def __str__(self):
        return self.packed

    @property
    def package(self) -> object:
        return eval(self.payload)

    @classmethod
    def from_string(cls, string: str):
        string = string.removeprefix("persistence_")
        tag, payload = string.split(":")
        return cls(tag, payload)

    @classmethod
    def new(cls, tag: str, package: "SupportsRepr"):
        try:
            tested = eval(repr(package))
            if type(tested) != type(package):
                raise ParseError(
                    "Evaluation the repr of the package did not return the same type."
                )
            if len(tag) + 1 + len(repr(package)) > 100:
                raise ParseError("The tag and payload combined are longer than 100 characters.")
            return cls(tag, repr(package))
        except:
            raise ParseError("The package could not be evaluated because it didn't properly implement __repr__.")


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="interactions-persistence",
    version="0.1.0",
    description="Encode objects in custom_ids.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Dworv/interactions-persistence",
    author="Dworv",
    author_email="dwarvyt@gmail.com",
    license="GNU",
    packages=["interactions.ext.persistence"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["discord-py-interactions"],
)


from inspect import getmembers, iscoroutinefunction
from types import MethodType
from interactions import Client, ComponentContext, CommandContext, Extension, extension_listener

from .decor import persistent_component, persistent_modal
from .parse import PersistentCustomID


class PersistenceExtension(Extension):
    def __new__(cls, client: Client, *args, **kwargs):

        self = super().__new__(cls, client, *args, **kwargs)

        for _, func in getmembers(self, predicate=iscoroutinefunction):
            if hasattr(func, "__persistence_type__"):
                if func.__persistence_type__ == "component": 
                    client.event(func, name="component_persistence_" + func.__persistence_tag__)
                elif func.__persistence_type__ == "modal":
                    client.event(func, name="modal_persistence_" + func.__persistence_tag__)

        return self

class Persistence(Extension):
    def __init__(self, bot: Client):
        bot.persistent_component = MethodType(persistent_component, bot)
        bot.persistent_modal = MethodType(persistent_modal, bot)

    @extension_listener(name="on_component")
    async def _on_component(self, ctx: ComponentContext):
        if not ctx.custom_id.startswith("persistence_"):
            return
        custom_id = PersistentCustomID.from_string(ctx.custom_id)
        listener = self.client._websocket._dispatch
        for name, funcs in listener.events.items():
            if name == "component_persistence_" + custom_id.tag:
                for func in funcs:
                    await func(ctx, custom_id.package)
                break

    @extension_listener(name="on_modal")
    async def _on_modal(self, ctx: CommandContext):
        if not ctx.data.custom_id.startswith("persistence_"):
            return
        custom_id = PersistentCustomID.from_string(ctx.data.custom_id)
        listener = self.client._websocket._dispatch
        answers = []
        if ctx.data._json.get("components"):
            for component in ctx.data.components:
                if component.get("components"):
                    answers.append(
                        [_value["value"] for _value in component["components"]][0]
                    )
                else:
                    answers.append([_value.value for _value in component.components][0])
        for name, funcs in listener.events.items():
            if name == "modal_persistence_" + custom_id.tag:
                for func in funcs:
                    await func(ctx, custom_id.package, *answers)
                break